@@ -83,7 +83,7 @@
 #include <afs/param.h>
 
 RCSID
-    ("$Header: /cvs/openafs/src/viced/callback.c,v 1.55.2.4 2004/11/09 17:17:49 shadow Exp $");
+    ("$Header: /cvs/openafs/src/viced/callback.c,v 1.55.2.18 2007/10/24 15:37:32 shadow Exp $");
 
 #include <stdio.h>
 #include <stdlib.h>		/* for malloc() */
@@ -328,7 +328,7 @@ FindFE(register AFSFid * fid)
     for (fei = HashTable[hash]; fei; fei = fe->fnext) {
 	fe = itofe(fei);
 	if (fe->volid == fid->Volume && fe->unique == fid->Unique
-	    && fe->vnode == fid->Vnode)
+	    && fe->vnode == fid->Vnode && (fe->status & FE_LATER) != FE_LATER)
 	    return fe;
     }
     return 0;
@@ -697,6 +697,7 @@ AddCallBack1_r(struct host *host, AFSFid * fid, afs_uint32 * thead, int type,
 	fe->vnode = fid->Vnode;
 	fe->unique = fid->Unique;
 	fe->ncbs = 0;
+	fe->status = 0;
 	hash = VHash(fid->Volume, fid->Unique);
 	fe->fnext = HashTable[hash];
 	HashTable[hash] = fetoi(fe);
@@ -720,6 +721,11 @@ AddCallBack1_r(struct host *host, AFSFid * fid, afs_uint32 * thead, int type,
 	    TDel(cb);
 	    TAdd(cb, Thead);
 	}
+	if (newfe == NULL) {    /* we are using the new FE */
+            fe->firstcb = cbtoi(cb);
+            fe->ncbs++;
+            cb->fhead = fetoi(fe);
+        }
     } else {
 	cb = newcb;
 	newcb = NULL;
@@ -772,6 +778,7 @@ MultiBreakCallBack_r(struct cbstruct cba[], int ncbas,
     int i, j;
     struct rx_connection *conns[MAX_CB_HOSTS];
     static struct AFSCBs tc = { 0, 0 };
+    int multi_to_cba_map[MAX_CB_HOSTS];
 
     assert(ncbas <= MAX_CB_HOSTS);
 
@@ -782,6 +789,7 @@ MultiBreakCallBack_r(struct cbstruct cba[], int ncbas,
 	    continue;
 	}
 	rx_GetConnection(thishost->callback_rxcon);
+	multi_to_cba_map[j] = i;
 	conns[j++] = thishost->callback_rxcon;
 
 #ifdef	ADAPT_MTU
@@ -800,28 +808,14 @@ MultiBreakCallBack_r(struct cbstruct cba[], int ncbas,
 		struct host *hp;
 		char hoststr[16];
 
-		idx = 0;
-		/* If there's an error, we have to hunt for the right host. 
-		 * The conns array _should_ correspond one-to-one to the cba
-		 * array, except in some rare cases it might be missing one 
-		 * or more elements.  So the optimistic case is almost 
-		 * always right.  At worst, it's the starting point for the 
-		 * hunt. */
-		for (hp = 0, i = multi_i; i < j; i++) {
-		    hp = cba[i].hp;	/* optimistic, but usually right */
-		    if (!hp) {
-			break;
-		    }
-		    if (conns[multi_i] == hp->callback_rxcon) {
-			idx = cba[i].thead;
-			break;
-		    }
-		}
+		i = multi_to_cba_map[multi_i];
+		hp = cba[i].hp;
+		idx = cba[i].thead;
 
-		if (!hp) {
+		if (!hp || !idx) {
 		    ViceLog(0,
-			    ("BCB: INTERNAL ERROR: hp=%x, cba=%x\n", hp,
-			     cba));
+			    ("BCB: INTERNAL ERROR: hp=%x, cba=%x, thead=%u\n", 
+			     hp, cba, idx));
 		} else {
 		    /* 
 		     ** try breaking callbacks on alternate interface addresses
@@ -829,7 +823,7 @@ MultiBreakCallBack_r(struct cbstruct cba[], int ncbas,
 		    if (MultiBreakCallBackAlternateAddress(hp, afidp)) {
 			if (ShowProblems) {
 			    ViceLog(7,
-				    ("BCB: Failed on file %u.%u.%u, host %s:%d is down\n",
+				    ("BCB: Failed on file %u.%u.%u, Host %s:%d is down\n",
 				     afidp->AFSCBFids_val->Volume,
 				     afidp->AFSCBFids_val->Vnode,
 				     afidp->AFSCBFids_val->Unique,
@@ -1094,7 +1088,7 @@ BreakDelayedCallBacks_r(struct host *host)
 
     cbstuff.nbreakers++;
     if (!(host->hostFlags & RESETDONE) && !(host->hostFlags & HOSTDELETED)) {
-	host->hostFlags &= ~ALTADDR;	/* alterrnate addresses are invalid */
+	host->hostFlags &= ~ALTADDR;	/* alternate addresses are invalid */
 	cb_conn = host->callback_rxcon;
 	rx_GetConnection(cb_conn);
 	if (host->interface) {
@@ -1112,7 +1106,7 @@ BreakDelayedCallBacks_r(struct host *host)
 	if (code) {
 	    if (ShowProblems) {
 		ViceLog(0,
-			("CB: Call back connect back failed (in break delayed) for %s:%d\n",
+			("CB: Call back connect back failed (in break delayed) for Host %s:%d\n",
 			 afs_inet_ntoa_r(host->host, hoststr),
 			 ntohs(host->port)));
 	    }
@@ -1159,7 +1153,7 @@ BreakDelayedCallBacks_r(struct host *host)
 		int i;
 		if (ShowProblems) {
 		    ViceLog(0,
-			    ("CB: XCallBackBulk failed, host=%s:%d; callback list follows:\n",
+			    ("CB: XCallBackBulk failed, Host %s:%d; callback list follows:\n",
 			     afs_inet_ntoa_r(host->host, hoststr),
 			     ntohs(host->port)));
 		}
@@ -1187,7 +1181,7 @@ BreakDelayedCallBacks_r(struct host *host)
 
     cbstuff.nbreakers--;
     /* If we succeeded it's always ok to unset HFE_LATER */
-    if (!host->hostFlags & VENUSDOWN)
+    if (!(host->hostFlags & VENUSDOWN))
 	host->hostFlags &= ~HFE_LATER;
     return (host->hostFlags & VENUSDOWN);
 }
@@ -1214,11 +1208,11 @@ MultiBreakVolumeCallBack_r(struct host *host, int isheld,
 	    return 0;		/* Release hold */
 	}
 	ViceLog(8,
-		("BVCB: volume call back for host %s:%d failed\n",
+		("BVCB: volume call back for Host %s:%d failed\n",
 		 afs_inet_ntoa_r(host->host, hoststr), ntohs(host->port)));
 	if (ShowProblems) {
 	    ViceLog(0,
-		    ("CB: volume callback for host %s:%d failed\n",
+		    ("CB: volume callback for Host %s:%d failed\n",
 		     afs_inet_ntoa_r(host->host, hoststr),
 		     ntohs(host->port)));
 	}
@@ -1402,7 +1396,9 @@ BreakVolumeCallBacksLater(afs_uint32 volume)
 
     ViceLog(25, ("Fsync thread wakeup\n"));
 #ifdef AFS_PTHREAD_ENV
+    FSYNC_LOCK;
     assert(pthread_cond_broadcast(&fsync_cond) == 0);
+    FSYNC_UNLOCK;
 #else
     LWP_NoYieldSignal(fsync_wait);
 #endif
@@ -1421,6 +1417,7 @@ BreakLaterCallBacks(void)
     struct host *host;
     struct VCBParams henumParms;
     unsigned short tthead = 0;	/* zero is illegal value */
+    char hoststr[16];
 
     /* Unchain first */
     ViceLog(25, ("Looking for FileEntries to unchain\n"));
@@ -1460,16 +1457,23 @@ BreakLaterCallBacks(void)
     for (fe = myfe; fe;) {
 	register struct CallBack *cbnext;
 	for (cb = itocb(fe->firstcb); cb; cb = cbnext) {
-	    host = h_itoh(cb->hhead);
-	    h_Hold_r(host);
 	    cbnext = itocb(cb->cnext);
-	    if (!tthead || (TNorm(tthead) < TNorm(cb->thead))) {
-		tthead = cb->thead;
+	    host = h_itoh(cb->hhead);
+	    if (cb->status == CB_DELAYED) {
+		h_Hold_r(host);
+		if (!tthead || (TNorm(tthead) < TNorm(cb->thead))) {
+		    tthead = cb->thead;
+		}
+		TDel(cb);
+		HDel(cb);
+		CDel(cb, 0);	/* Don't let CDel clean up the fe */
+		/* leave hold for MultiBreakVolumeCallBack to clear */
+	    } else {
+		ViceLog(125,
+			("Found host %s:%d non-DELAYED cb for %u:%u:%u\n", 
+			 afs_inet_ntoa_r(host->host, hoststr),
+			 ntohs(host->port), fe->vnode, fe->unique, fe->volid));
 	    }
-	    TDel(cb);
-	    HDel(cb);
-	    CDel(cb, 0);	/* Don't let CDel clean up the fe */
-	    /* leave hold for MultiBreakVolumeCallBack to clear */
 	}
 	myfe = fe;
 	fe = (struct FileEntry *)((struct object *)fe)->next;
@@ -1552,34 +1556,88 @@ CleanupTimedOutCallBacks_r(void)
 }
 
 static struct host *lih_host;
-static int lih_host_held = 0;
+static int lih_host_held;
 
 static int
 lih_r(register struct host *host, register int held,
       register struct host *hostp)
 {
-    lih_host_held = 0;
     if (host->cblist
 	&& ((hostp && host != hostp) || (!held && !h_OtherHolds_r(host)))
 	&& (!lih_host || host->ActiveCall < lih_host->ActiveCall)) {
+	if (lih_host != NULL && lih_host_held) {
+	    h_Release_r(lih_host);
+	}
 	lih_host = host;
+	lih_host_held = !held;
+	held = 1;
     }
-    if (!held) {
+    return held;
+}
+
+/* This version does not allow 'host' to be selected unless its ActiveCall 
+ * is newer than 'hostp' which is the host with the oldest ActiveCall from
+ * the last pass (if it is provided).  We filter out any hosts that are
+ * are held by other threads.
+ */
+static int
+lih0_r(register struct host *host, register int held,
+      register struct host *hostp)
+{
+    if (host->cblist
+	&& (hostp && host != hostp) 
+	&& (!held && !h_OtherHolds_r(host))
+	&& (!lih_host || host->ActiveCall < lih_host->ActiveCall) 
+	&& (!hostp || host->ActiveCall > hostp->ActiveCall)) {
+	if (lih_host != NULL && lih_host_held) {
+	    h_Release_r(lih_host);
+	}
+	lih_host = host;
+	lih_host_held = !held;
+	held = 1;
+    }
+    return held;
+}
+
+/* This version does not allow 'host' to be selected unless its ActiveCall 
+ * is newer than 'hostp' which is the host with the oldest ActiveCall from
+ * the last pass (if it is provided).  In this second varient, we do not 
+ * prevent held hosts from being selected.
+ */
+static int
+lih1_r(register struct host *host, register int held,
+      register struct host *hostp)
+{
+    if (host->cblist
+	&& (hostp && host != hostp) 
+	&& (!lih_host || host->ActiveCall < lih_host->ActiveCall) 
+	&& (!hostp || host->ActiveCall > hostp->ActiveCall)) {
+	if (lih_host != NULL && lih_host_held) {
+	    h_Release_r(lih_host);
+	}
+	lih_host = host;
+	lih_host_held = !held;
 	held = 1;
-	lih_host_held = 1;
     }
     return held;
 }
 
 /* This could be upgraded to get more space each time */
-/* first pass: find the oldest host which isn't held by anyone */
-/* second pass: find the oldest host who isn't "me" */
+/* first pass: sequentially find the oldest host which isn't held by
+               anyone for which we can clear callbacks;
+	       skipping 'hostp' */
+/* second pass: sequentially find the oldest host regardless of 
+               whether or not the host is held; skipping 'hostp' */
+/* third pass: attempt to clear callbacks from 'hostp' */
 /* always called with hostp unlocked */
+
+/* Note: hostlist is ordered most recently created host first and 
+ * its order has no relationship to the most recently used. */
 extern struct host *hostList;
 static int
 GetSomeSpace_r(struct host *hostp, int locked)
 {
-    register struct host *hp, *hp1 = (struct host *)0, *hp2 = hostList;
+    register struct host *hp, *hp1, *hp2;
     int i = 0;
 
     cbstuff.GotSomeSpaces++;
@@ -1589,27 +1647,28 @@ GetSomeSpace_r(struct host *hostp, int locked)
 	cbstuff.GSS3++;
 	return 0;
     }
+
+    i = 0;
+    hp1 = NULL;
+    hp2 = hostList;
     do {
 	lih_host = 0;
-	h_Enumerate_r(lih_r, hp2, (char *)hp1);
+	h_Enumerate_r(i == 0 ? lih0_r : lih1_r, hp2, (char *)hp1);
 	hp = lih_host;
 	if (hp) {
+	    /* set in lih_r! private copy before giving up H_LOCK */
+	    int lih_host_held2=lih_host_held;   
 	    cbstuff.GSS4++;
-	    if (!ClearHostCallbacks_r(hp, 0 /* not locked or held */ )) {
-		if (lih_host_held) 
+	    if ((hp != hostp) && !ClearHostCallbacks_r(hp, 0 /* not locked or held */ )) {
+		if (lih_host_held2)
 		    h_Release_r(hp);
 		return 0;
 	    }
-	    if (lih_host_held) 
+	    if (lih_host_held2)
 		h_Release_r(hp);
-	    hp2 = hp->next;
-	} else {
+	    hp1 = hp;
 	    hp2 = hostList;
-	    hp1 = hostp;
-	    cbstuff.GSS1++;
-	    ViceLog(5,
-		    ("GSS: Try harder for longest inactive host cnt= %d\n",
-		     i));
+	} else {
 	    /*
 	     * Next time try getting callbacks from any host even if
 	     * it's deleted (that's actually great since we can freely
@@ -1618,13 +1677,16 @@ GetSomeSpace_r(struct host *hostp, int locked)
 	     * callback timeout arrives).
 	     */
 	    i++;
+	    hp1 = NULL;
+	    hp2 = hostList;
+	    cbstuff.GSS1++;
+	    ViceLog(5,
+		    ("GSS: Try harder for longest inactive host cnt= %d\n",
+		     i));
 	}
     } while (i < 2);
-    /*
-     * No choice to clear this host's callback state
-     */
-    /* third pass: we still haven't gotten any space, so we free what we had
-     * previously passed over. */
+
+    /* Could not obtain space from other hosts, clear hostp's callback state */
     cbstuff.GSS2++;
     if (!locked) {
 	h_Lock_r(hostp);
@@ -1968,7 +2030,7 @@ MultiBreakCallBackAlternateAddress_r(struct host *host,
     int i, j;
     struct rx_connection **conns;
     struct rx_connection *connSuccess = 0;
-    afs_int32 *addr;
+    struct AddrPort *interfaces;
     static struct rx_securityClass *sc = 0;
     static struct AFSCBs tc = { 0, 0 };
     char hoststr[16];
@@ -1977,10 +2039,8 @@ MultiBreakCallBackAlternateAddress_r(struct host *host,
     if (!host->interface)
 	return 1;		/* failure */
 
-    assert(host->interface->numberOfInterfaces > 0);
-
     /* the only address is the primary interface */
-    if (host->interface->numberOfInterfaces == 1)
+    if (host->interface->numberOfInterfaces <= 1)
 	return 1;		/* failure */
 
     /* initialise a security object only once */
@@ -1988,9 +2048,9 @@ MultiBreakCallBackAlternateAddress_r(struct host *host,
 	sc = rxnull_NewClientSecurityObject();
 
     i = host->interface->numberOfInterfaces;
-    addr = calloc(i, sizeof(afs_int32));
+    interfaces = calloc(i, sizeof(struct AddrPort));
     conns = calloc(i, sizeof(struct rx_connection *));
-    if (!addr || !conns) {
+    if (!interfaces || !conns) {
 	ViceLog(0,
 		("Failed malloc in MultiBreakCallBackAlternateAddress_r\n"));
 	assert(0);
@@ -1999,12 +2059,14 @@ MultiBreakCallBackAlternateAddress_r(struct host *host,
     /* initialize alternate rx connections */
     for (i = 0, j = 0; i < host->interface->numberOfInterfaces; i++) {
 	/* this is the current primary address */
-	if (host->host == host->interface->addr[i])
+	if (host->host == host->interface->interface[i].addr &&
+	    host->port == host->interface->interface[i].port)
 	    continue;
 
-	addr[j] = host->interface->addr[i];
+	interfaces[j] = host->interface->interface[i];
 	conns[j] =
-	    rx_NewConnection(host->interface->addr[i], host->port, 1, sc, 0);
+	    rx_NewConnection(interfaces[j].addr, 
+			     interfaces[j].port, 1, sc, 0);
 	rx_SetConnDeadTime(conns[j], 2);
 	rx_SetConnHardDeadTime(conns[j], AFS_HARDDEADTIME);
 	j++;
@@ -2023,13 +2085,16 @@ MultiBreakCallBackAlternateAddress_r(struct host *host,
 	    if (host->callback_rxcon)
 		rx_DestroyConnection(host->callback_rxcon);
 	    host->callback_rxcon = conns[multi_i];
-	    host->host = addr[multi_i];
+            hashDelete_r(host->host, host->port, host);
+	    host->host = interfaces[multi_i].addr;
+	    host->port = interfaces[multi_i].port;
+            hashInsert_r(host->host, host->port, host);
 	    connSuccess = conns[multi_i];
 	    rx_SetConnDeadTime(host->callback_rxcon, 50);
 	    rx_SetConnHardDeadTime(host->callback_rxcon, AFS_HARDDEADTIME);
 	    ViceLog(125,
 		    ("multibreakcall success with addr %s\n",
-		     afs_inet_ntoa_r(addr[multi_i], hoststr)));
+		     afs_inet_ntoa_r(interfaces[multi_i].addr, hoststr)));
 	    H_UNLOCK;
 	    multi_Abort;
 	}
@@ -2041,7 +2106,7 @@ MultiBreakCallBackAlternateAddress_r(struct host *host,
 	if (conns[i] != connSuccess)
 	    rx_DestroyConnection(conns[i]);
 
-    free(addr);
+    free(interfaces);
     free(conns);
 
     if (connSuccess)
@@ -2052,8 +2117,8 @@ MultiBreakCallBackAlternateAddress_r(struct host *host,
 
 
 /*
-** try multiRX probes to host. 
-** return 0 on success, non-zero on failure
+** try multi_RX probes to host. 
+** return 0 on success, non-0 on failure
 */
 int
 MultiProbeAlternateAddress_r(struct host *host)
@@ -2061,7 +2126,7 @@ MultiProbeAlternateAddress_r(struct host *host)
     int i, j;
     struct rx_connection **conns;
     struct rx_connection *connSuccess = 0;
-    afs_int32 *addr;
+    struct AddrPort *interfaces;
     static struct rx_securityClass *sc = 0;
     char hoststr[16];
 
@@ -2069,20 +2134,18 @@ MultiProbeAlternateAddress_r(struct host *host)
     if (!host->interface)
 	return 1;		/* failure */
 
-    assert(host->interface->numberOfInterfaces > 0);
-
     /* the only address is the primary interface */
-    if (host->interface->numberOfInterfaces == 1)
-	return 1;		/* failure */
+    if (host->interface->numberOfInterfaces <= 1)
+        return 1;               /* failure */
 
     /* initialise a security object only once */
     if (!sc)
 	sc = rxnull_NewClientSecurityObject();
 
     i = host->interface->numberOfInterfaces;
-    addr = calloc(i, sizeof(afs_int32));
+    interfaces = calloc(i, sizeof(struct AddrPort));
     conns = calloc(i, sizeof(struct rx_connection *));
-    if (!addr || !conns) {
+    if (!interfaces || !conns) {
 	ViceLog(0, ("Failed malloc in MultiProbeAlternateAddress_r\n"));
 	assert(0);
     }
@@ -2090,12 +2153,14 @@ MultiProbeAlternateAddress_r(struct host *host)
     /* initialize alternate rx connections */
     for (i = 0, j = 0; i < host->interface->numberOfInterfaces; i++) {
 	/* this is the current primary address */
-	if (host->host == host->interface->addr[i])
+	if (host->host == host->interface->interface[i].addr &&
+	    host->port == host->interface->interface[i].port)
 	    continue;
 
-	addr[j] = host->interface->addr[i];
+	interfaces[j] = host->interface->interface[i];
 	conns[j] =
-	    rx_NewConnection(host->interface->addr[i], host->port, 1, sc, 0);
+	    rx_NewConnection(interfaces[j].addr, 
+			     interfaces[j].port, 1, sc, 0);
 	rx_SetConnDeadTime(conns[j], 2);
 	rx_SetConnHardDeadTime(conns[j], AFS_HARDDEADTIME);
 	j++;
@@ -2113,17 +2178,40 @@ MultiProbeAlternateAddress_r(struct host *host)
 	    H_LOCK;
 	    if (host->callback_rxcon)
 		rx_DestroyConnection(host->callback_rxcon);
+            hashDelete_r(host->host, host->port, host);
 	    host->callback_rxcon = conns[multi_i];
-	    host->host = addr[multi_i];
+	    host->host = interfaces[multi_i].addr;
+	    host->port = interfaces[multi_i].port;
+            hashInsert_r(host->host, host->port, host);
 	    connSuccess = conns[multi_i];
 	    rx_SetConnDeadTime(host->callback_rxcon, 50);
 	    rx_SetConnHardDeadTime(host->callback_rxcon, AFS_HARDDEADTIME);
 	    ViceLog(125,
-		    ("multiprobe success with addr %s\n",
-		     afs_inet_ntoa_r(addr[multi_i], hoststr)));
+		    ("multiprobe success for host %x; new primary addr %s:%d\n",
+		     host, afs_inet_ntoa_r(host->host, hoststr),
+                     ntohs(host->port)));
 	    H_UNLOCK;
 	    multi_Abort;
-	}
+	} else {
+	    ViceLog(125,
+		    ("multiprobe failure with addr %s:%d\n",
+		     afs_inet_ntoa_r(interfaces[multi_i].addr, hoststr),
+                     ntohs(interfaces[multi_i].port)));
+            
+            /* This is less than desirable but its the best we can do.
+             * The AFS Cache Manager will return either 0 for a Uuid  
+             * match and a 1 for a non-match.   If the error is 1 we 
+             * therefore know that our mapping of IP address to Uuid 
+             * is wrong.   We should attempt to find the correct
+             * Uuid and fix the host tables.
+             */
+            if (multi_error == 1) {
+                /* remove the current alternate address from this host */
+                H_LOCK;
+                removeInterfaceAddr_r(host, interfaces[multi_i].addr, interfaces[multi_i].port);
+                H_UNLOCK;
+            }
+        }
     }
     multi_End_Ignore;
     H_LOCK;
@@ -2132,7 +2220,7 @@ MultiProbeAlternateAddress_r(struct host *host)
 	if (conns[i] != connSuccess)
 	    rx_DestroyConnection(conns[i]);
 
-    free(addr);
+    free(interfaces);
     free(conns);
 
     if (connSuccess)
